import '../-internals/meta/lib/meta.js';
import '../../shared-chunks/mandatory-setter-CsT0mMvz.js';
import '@embroider/macros';
import '../../shared-chunks/index-C_0RkUaC.js';
import '../../@glimmer/destroyable/index.js';
import '../../@glimmer/validator/index.js';
import { z as addListener, I as sendEvent, A as removeListener, J as hasListeners } from '../../shared-chunks/cache-DfGfGjt9.js';
export { K as on } from '../../shared-chunks/cache-DfGfGjt9.js';
import '../../@glimmer/manager/index.js';
import '../../shared-chunks/env-BJLX2Arx.js';
import Mixin from './mixin.js';

/**
@module @ember/object/evented
*/

/**
  This mixin allows for Ember objects to subscribe to and emit events.

  ```app/utils/person.js
  import EmberObject from '@ember/object';
  import Evented from '@ember/object/evented';

  export default EmberObject.extend(Evented, {
    greet() {
      // ...
      this.trigger('greet');
    }
  });
  ```

  ```javascript
  var person = Person.create();

  person.on('greet', function() {
    console.log('Our person has greeted');
  });

  person.greet();

  // outputs: 'Our person has greeted'
  ```

  You can also chain multiple event subscriptions:

  ```javascript
  person.on('greet', function() {
    console.log('Our person has greeted');
  }).one('greet', function() {
    console.log('Offer one-time special');
  }).off('event', this, forgetThis);
  ```

  @class Evented
  @public
 */

const Evented = Mixin.create({
  on(name, target, method) {
    addListener(this, name, target, method);
    return this;
  },
  one(name, target, method) {
    addListener(this, name, target, method, true);
    return this;
  },
  trigger(name, ...args) {
    sendEvent(this, name, args);
  },
  off(name, target, method) {
    removeListener(this, name, target, method);
    return this;
  },
  has(name) {
    return hasListeners(this, name);
  }
});

export { Evented as default };
